# Copyright (c) 2020 Edwin Wise
# MIT License
# See LICENSE for details
"""
    Controls are widgets that are manipulated and return a value. They may
    update their function on move and release.
"""
import logging
import math

import pygame
from custom_inherit import doc_inherit

from pi_touch_gui._widget_bases import IControlWidget

LOG = logging.getLogger(__name__)


class Dial(IControlWidget):
    """ Circular input control dial.
    """

    @doc_inherit(IControlWidget.__init__, style='numpy_with_merge')
    def __init__(self, position, size, **kwargs):
        """ Initialize the dial.
        """
        # Dials don't use label, so block it
        if kwargs.get('label') is not None:
            raise ValueError("Dials don't use their value as the label.")

        diameter = min(size[0], size[1])
        radius = diameter >> 1
        kw_font_size = kwargs.pop('font_size', None)
        font_size = (radius >> 1) if kw_font_size is None else kw_font_size

        # The parent init sets convenience properties like self.x, etc
        super(Dial, self).__init__(position, (diameter, diameter),
                                   font_size=font_size,
                                   **kwargs)
        self.radius = radius
        self.cx, self.cy = (self.x + self.w // 2), (self.y + self.h // 2)

    def touch_inside(self, touch):
        """ The interior of the dial is determined by a circle.
        """
        x, y = touch.position
        distance = math.hypot(x - self.cx, y - self.cy)
        if distance <= self.radius:
            return True

        return False

    def on_move(self, touch):
        # If too many touches, we bail
        if len(self.touches) > 1:
            return

        dx = touch.x - self.cx
        dy = touch.y - self.cy

        # Value is from 0.0 to 1.0 (percent of range)
        self.value = (math.atan2(dy, dx) % math.tau) / math.tau
        self.adjusted_value = self.map_value_to_adjusted_value(self.value)

        super(Dial, self).on_move(touch)

    def render(self, surface):
        wedge_rect = (self.x, self.y, self.w, self.h)
        handle_pos = (
            int(self.cx + (self.radius * math.cos(self.value * math.tau))),
            int(self.cy + (self.radius * math.sin(self.value * math.tau))))

        center = (self.cx, self.cy)
        # Draw the filled overall dial
        pygame.draw.circle(surface, self.color1, center, self.radius, 0)
        # Draw the active pie-wedge
        pygame.draw.arc(surface, self.color2, wedge_rect,
                        -self.value * math.tau, 0.0,
                        int(self.radius * 0.5))
        # Draw the handle line
        pygame.draw.line(surface, self.label_color1, center, handle_pos,
                         3)
        # Draw the handle circle
        pygame.draw.circle(surface, self.label_color1, handle_pos,
                           int(self.radius * 0.1), 0)

        self.render_centered_text(surface, self.value_str(),
                                  self.label_color1, self.color1)

        if self._focus:
            self.render_focus(surface)


class Slider(IControlWidget):
    """ Variable value input linear slider
    """

    @doc_inherit(IControlWidget.__init__, style='numpy_with_merge')
    def __init__(self, position, size, **kwargs):
        """ Initialize the slider.

        color1 is the slider background fill
        color2 is the slider outline and active value fill
        label_color1 is the slider line (and label)
        """
        # Sliders don't use label, so block it
        if kwargs.get('label') is not None:
            raise ValueError("Dials don't use their value as the label.")

        super(Slider, self).__init__(position, size, **kwargs)

    def on_move(self, touch):
        # If too many touches, bail
        if len(self.touches) > 1:
            return

        x, y = touch.position

        # Compute X/Y relative to the slider
        x -= self.x
        y -= self.y

        # Check our ratios...
        if self.w > self.h:
            # Horizontal Slider
            if 0 <= x <= self.w:
                self.value = float(x) / float(self.w)

        elif self.h > self.w:
            # Vertical Slider
            if 0 <= y <= self.h:
                self.value = (self.h - float(y)) / float(self.h)

        self.adjusted_value = self.map_value_to_adjusted_value(self.value)

        super(Slider, self).on_move(touch)

    def render(self, surface):
        pygame.draw.rect(surface, self.color1, (self.position, self.size), 0)

        if self.w > self.h:
            # Horizontal slider
            value_w = int(self.w * self.value)
            value_x = self.x + value_w
            pygame.draw.rect(surface, self.color2,
                             ((self.x, self.y), (value_w, self.h)))
            pygame.draw.line(surface, self.label_color1,
                             (value_x, self.y), (value_x, self.y + self.h),
                             3)

        elif self.h > self.w:
            # Vertical slider
            value_h = int(self.h * self.value)
            value_y = self.y + (self.h - value_h)
            pygame.draw.rect(surface, self.color2,
                             ((self.x, value_y), (self.w, value_h)))
            pygame.draw.line(surface, self.label_color1,
                             (self.x, value_y), (self.x + self.w, value_y),
                             3)

        pygame.draw.rect(surface, self.color2, (self.position, self.size),
                         3)

        self.render_centered_text(surface, self.value_str(), self.label_color1)

        if self._focus:
            self.render_focus(surface)



# Copyright (c) 2020 Edwin Wise
# MIT License
# See LICENSE for details
"""
    Test - Page with poweroff button
"""
import os

from colors import RED
from gui import Page
from tests._utilities import pulsing_light_func
from widget_buttons import Button
from widget_displays import Label, Light


def poweroff_page():
    def poweroff_button_func(button, pushed):
        os.system('sudo poweroff')
        return None

    _exit_label_1 = Label((0, 0), (800, 50), label="Nothing to see here")
    _exit_label_2 = Label((0, 50), (800, 430), font_size=100,
                          label_color1=RED, label="... move along!")

    _pulse1_light = Light((0, 0), (50, 50), function=pulsing_light_func)
    _pulse1_light.rate = 100

    _pulse2_light = Light((750, 0), (50, 50), function=pulsing_light_func)
    _pulse2_light.rate = 105

    _pulse3_light = Light((0, 430), (50, 50), function=pulsing_light_func)
    _pulse3_light.rate = 110

    _pulse4_light = Light((750, 430), (50, 50), pulsing_light_func)
    _pulse4_light.rate = 115

    _off_button = Button((200, 320), (400, 150),
                         style=Button.OBROUND,
                         label="-~= OFF =~-",
                         function=poweroff_button_func)

    def _configure_page(go_to_page):
        """ Custom configuration on the entry page - sets the page and
        button functions.

        Parameters
        ----------
        go_to_page : Callable[[IWidget, bool], Optional['Page']]
        """
        _poweroff_page.function = go_to_page

    _poweroff_page = Page(
        # Tests dynamic Light functions, Label, and default Page function
        [
            _exit_label_1, _exit_label_2,
            _pulse1_light, _pulse2_light, _pulse3_light, _pulse4_light,
            _off_button
        ],
    )
    _poweroff_page.config = _configure_page

    return _poweroff_page



# Copyright (c) 2020 Edwin Wise
# MIT License
# See LICENSE for details
"""
    Test - An entry page with pulsing lights
"""

# ---- Pulsing light test function (light)

from colors import RED
from gui import Page
from tests._utilities import pulsing_light_func
from widget_buttons import Button
from widget_displays import Light, Label


def entry_page():
    _entry_label = Label((0, 0), (800, 50), label="ENTER"),

    _pulse1_light = Light((0, 0), (50, 50), function=pulsing_light_func)
    _pulse1_light.rate = 400

    _pulse2_light = Light((750, 0), (50, 50), function=pulsing_light_func)
    _pulse2_light.rate = 300

    _pulse3_light = Light((0, 430), (50, 50), function=pulsing_light_func)
    _pulse3_light.rate = 200

    _pulse4_light = Light((750, 430), (50, 50), pulsing_light_func)
    _pulse4_light.rate = 100

    _touch_me_button = Button((200, 150), (400, 100), font_size=100,
                              label_color1=RED, label="Touch Me")

    def _configure_page(go_to_page):
        """ Custom configuration on the entry page - sets the page and
        button functions.

        Parameters
        ----------
        go_to_page : Callable[[IWidget, bool], Optional['Page']]
        """
        _entry_page.function = go_to_page
        _touch_me_button.function = go_to_page

    _entry_page = Page(
        # Tests dynamic Light functions, Label, and default Page function
        [
            _pulse1_light,
            _pulse2_light,
            _pulse3_light,
            _pulse4_light,
            _touch_me_button
        ]
    )
    # Add a custom attribute, which is monkey-patching at its worst
    _entry_page.config = _configure_page

    return _entry_page



# Copyright (c) 2020 Edwin Wise
# MIT License
# See LICENSE for details
"""
    Graphical User Interface objects

    A `Page` is a set of widgets that define the interaction and display
    surface of the GUI

    The `GUI` manages the hardware and runs the Pages
"""
import logging
import math
from typing import List, Optional, NoReturn

import pygame
from pygame.constants import *

from pi_touch_gui._utilities import backlog_error
from pi_touch_gui._widget_bases import IWidget, IWidgetInterface
from pi_touch_gui.colors import BLACK
from pi_touch_gui.multitouch.raspberrypi_ts import Touchscreen, TS_RELEASE

LOG = logging.getLogger(__name__)


class Page(object):
    WALL = 'wall'
    GO_WALL = -1
    GO_UP = 0
    GO_DOWN = 1
    GO_LEFT = 2
    GO_RIGHT = 3

    GO_NEIGHBORLESS = [-1, 1, -1, 1]
    GO_NAMES = ["Up", "Down", "Left", "Right"]

    def __init__(self, widgets=None, background=None, function=None):
        """ A collection of widgets that make up an interaction page.

        Parameters
        ----------
        widgets : List[IWidget]
        background : str
            Background image resource filepath for this Page
        function : Callable[[Page], Optional[Type['Page']]]
            Function to call if none of the widgets consume
            an event.
        """
        # If no widgets added on init, they can be added with `add_widgets()`
        # later.
        self._widgets = [] if widgets is None else widgets

        if background is not None:
            self._background = pygame.image.load(background).convert()
        else:
            self._background = None
        self.function = function
        self._focus_widget = None
        if widgets is None:
            self._widgets = []
            self._prev_focus_widget = None
        else:
            self._widgets = widgets
            self._prev_focus_widget = widgets[0]

        self._last_adjustment = pygame.time.get_ticks()
        self.adjustment_rate = 100

    @property
    def background(self):
        return self._background

    @background.setter
    def background(self, filepath):
        self._background = pygame.image.load(filepath).convert()

    def add_widgets(self, widgets):
        """ Add widgets to the Page.

        Parameters
        ----------
        widgets : List[IWidget]
        """
        self._widgets += widgets
        # Set or reset the base focus widget to the top
        self._prev_focus_widget = self._widgets[0]

    def find_widget(self, widget_or_id):
        """ Given a widget ID, find that widget.

        Parameters
        ----------
        widget_or_id : Union[IWidgetInterface, String]
            If a widget is passed in, return it directly. Otherwise search
            for the id and return the widget.

        Returns
        -------
        IWidgetInterface
        """
        # Deal with the WALL case, where we _know_ there isn't a widget
        if widget_or_id == Page.WALL:
            return None
        if not isinstance(widget_or_id, str):
            return widget_or_id

        return self._scan_for_widget_by_id(self._widgets, widget_or_id)

    def _scan_for_widget_by_id(self, widgets, widget_id):
        """ Recursively scan for a widget ID, returning the widget.

        Parameters
        ----------
        widgets : List[IWidgetInterface]
            A list of widgets, some of which may container lists of widgets
        widget_id : String
            The widget ID to search for

        Returns
        -------
        IWidgetInterface
        """
        for widget in widgets:
            # See if this is a widget container...
            found_widget = None
            try:
                # Assume the widget is a widget container and assert if not
                sub_widgets = widget._widgets
                found_widget = self._scan_for_widget_by_id(sub_widgets,
                                                           widget_id)
            except AttributeError:
                if widget.id == widget_id:
                    found_widget = widget
            if found_widget is not None:
                return found_widget
        return None

    def find_next_widget(self, widget_or_id, direction):
        """ Given a widget ID, find the widget AFTER that widget.

        Parameters
        ----------
        widget_or_id : Union[IWidgetInterface, String]
        direction : Integer
            Direction of scan, sign(direction)

        Returns
        -------
        IWidgetInterface
        """
        # Deal with the WALL case, where we _know_ there isn't a widget
        if widget_or_id == Page.WALL:
            return None
        try:
            widget_or_id = widget_or_id.id
        except AttributeError:
            pass

        return self._scan_for_next_widget_by_id(self._widgets,
                                                widget_or_id,
                                                math.copysign(1, direction))

    def _scan_for_next_widget_by_id(self, widgets, widget_id, step):
        """ Recursively scan for a widget ID, returning the widget AFTER the
        widget found.

        Parameters
        ----------
        widgets : List[IWidgetInterface]
            A list of widgets, some of which may container lists of widgets
        widget_id : String
            The widget ID to search for
        step : Integer
            The step (+1, -1) we are scanning next for.

        Returns
        -------
        IWidgetInterface
        """
        step = int(step)
        for idx, widget in enumerate(widgets):
            # See if this is a widget container...
            found_widget = None
            try:
                # Assume the widget is a widget container and assert if not
                sub_widgets = widget._widgets
                found_widget = self._scan_for_widget_by_id(sub_widgets,
                                                           widget_id)
            except AttributeError:
                if widget.id == widget_id:
                    try:
                        found_widget = widgets[idx + step]
                    except Exception as e:
                        return None
            if found_widget is not None:
                return found_widget
        return None

    def touch_handler(self, event, touch) -> Optional['Page']:
        """ Update all widgets with a touch event.

            The first widget to consume the touch wins, and downstream
            widgets won't be tested.

            If no widget consumes the event, we call `function(self) -> Page`

        Parameters
        ----------
        event : int
            Event code defined for the interface; e.g. _widget_bases.py imports
            TS_PRESS, TS_RELEASE, and TS_MOVE
        touch : Touch
            The Touch object for the event.

        Returns
        -------
        Optional[Page]
            If the widget function names a new page, it is returned here
        """
        for widget in self._widgets:
            consumed, new_page = widget.event(event, touch)
            if consumed:
                return new_page

        if event == TS_RELEASE and self.function:
            return self.function(self, False)
        return None

    def render(self, surface: pygame.Surface):
        """Redraw all widgets to screen.

        Call `render()` for all widgets in the Page, once per frame.

        Parameters
        ----------
        surface : pygame.Surface
            Surface (screen) to draw to
        """
        if self.background is not None:
            surface.blit(self.background, (0, 0))
        else:
            surface.fill(BLACK)
        for widget in self._widgets:
            try:
                widget.render(surface)
            except Exception as e:
                backlog_error(e, f"Problem rendering widget {widget}")
                raise

    def focus_widget(self, enable):
        """ Wake up the focus system for gesture control.

        Will always restart on the last widget that had focus.

        Parameters
        ----------
        enable : Bool
            If True, then focus; else de-focus
        """
        if self.is_focused() == enable:
            return

        prev_widget = self._prev_focus_widget
        self._focus_widget = self._prev_focus_widget
        if enable:
            if not prev_widget.can_focus():
                # This will walk the down chain until it finds a widget
                # that can take focus.  Skips starting labels and stuff.
                self.shift_widget_focus(Page.GO_DOWN)
            else:
                prev_widget.focus(True)
        else:
            self._focus_widget = None
            prev_widget.focus(False)

    def is_focused(self):
        """ Determine if we are in widget focus mode.

        Returns
        -------
        Bool
        """
        return self._focus_widget is not None

    def shift_widget_focus(self, direction):
        """ Shift focus one way or the other, wrapping at the boundaries.

        Parameters
        ----------
        direction : int
            Direction of motion; Page.WALL (no motion),
        """
        if direction == Page.GO_WALL:
            return
        if not self.is_focused():
            return

        start_widget = test_widget = self._focus_widget
        while True:
            next_widget = test_widget.neighbors[direction]

            # We have an IWidgetInterface, ID string, or 'wall'
            if next_widget == Page.WALL:
                # Hit a wall
                return
            elif next_widget is None:
                # No neighbor specified, so use a generic direction
                next_widget = self.find_next_widget(
                    test_widget,
                    Page.GO_NEIGHBORLESS[direction])
            else:
                # Neighbor as a widget ID or index
                next_widget = self.find_widget(next_widget)

            if test_widget.label == 'group':
                pause = True

            # In case we instantiated a neighbor, store it
            test_widget.neighbors[direction] = next_widget

            if next_widget == Page.GO_WALL:
                return
            elif next_widget == start_widget:
                return
            elif next_widget is None:
                return

            # Take the tentative next widget and make it current
            test_widget = next_widget
            if test_widget.can_focus():
                break

        try:
            if self._focus_widget is not None:
                self._focus_widget.focus(False)
            next_widget.focus(True)
        except Exception:
            pause = True
        self._focus_widget = next_widget
        self._prev_focus_widget = next_widget

    def adjust_widget(self, direction):
        """ Send an up/down value adjustment to the control widget.
        """
        time = pygame.time.get_ticks()
        if time - self._last_adjustment < self.adjustment_rate:
            return

        if self._focus_widget.adjust is not None:
            self._focus_widget.adjust(direction)
            self._last_adjustment = time

    def press_widget(self):
        """ Activate the function of the widget.
        """
        if self._focus_widget.press is not None:
            self._focus_widget.press()

    def release_widget(self):
        """ Activate the function of the widget.
        """
        if self._focus_widget.release is not None:
            return self._focus_widget.release()
        return None

    def iter_widgets(self):
        """ Iterate over the widgets in the page.

        Returns
        -------
        iter(IWidget)
        """
        return iter(self._widgets)


class GUI(object):
    # Constants that define the GUI behavior
    FPS = 60
    FADE_TIME = 60
    BLACKOUT_TIME = 300
    BRIGHT_LEVEL = 128
    FADE_LEVEL = 16

    # Initialize the event timer
    last_event = pygame.time.get_ticks()

    def __init__(self, touchscreen, bright=None):
        """ A Graphical User Interface for using a touchscreen.

        Parameters
        ----------
        touchscreen : Touchscreen
            See for example the `multitouch.raspberrypi_ts.Touchscreen`
        bright : int
            Brightness level (to balance visibility and screen wear).  Defaults
            to GUI.BRIGHT_LEVEL.
        """
        self._bright_level = self.BRIGHT_LEVEL if bright is None else bright
        self._touchscreen = touchscreen
        self._current_page = None
        self._first_page = None
        self._running = False
        self._faded = False
        self._blacked = False

        with open('/sys/class/backlight/rpi_backlight/max_brightness',
                  'r') as f:
            max_str = f.readline()
        self.max_brightness = min(int(max_str), 255)
        self.set_brightness(self._bright_level)

    def reset_display(self):
        """ Enable on the backlight and bring to the default brightness.
        """
        self.last_event = pygame.time.get_ticks()

        if self._faded:
            self.set_brightness(self._bright_level)
            self.fade = False
        if self._blacked:
            self.set_light(True)
            self._blacked = False

    def set_brightness(self, level):
        """ Set the backlight to a determined brightness level.

        Parameters
        ----------
        level : int
        """
        level_str = f"{min(self.max_brightness, level)}\n"
        with open('/sys/class/backlight/rpi_backlight/brightness', 'w') as f:
            f.write(level_str)

    def set_light(self, on):
        """ Turn the backlight activateion to On or Off

        Parameters
        ----------
        on : bool
            True for On, False for Off
        """
        if on is True:
            blank_str = "0"
        else:
            blank_str = "1"
            self._current_page = self._first_page
            self._current_page.render(self._touchscreen.surface)
        with open('/sys/class/backlight/rpi_backlight/bl_power', 'w') as f:
            f.write(blank_str)

    def run(self, page) -> NoReturn:
        """ Run the GUI, staring with the initial Page.

        Does not return, but does an exit() when done. Wrapping
        code in this thread needs to embed this in a Try/Finally to do cleanup.

        Parameters
        ----------
        page : Page
        """
        self._current_page = page
        self._first_page = page
        self._running = True

        self._touchscreen.run()

        fpsClock = pygame.time.Clock()

        def handle_touches(e, t):
            """ Call the `Page.touch_handler()` to process an event. """
            if self._blacked:
                # Consume touch if we are fully blacked out
                self.reset_display()
                return
            # Reset the brightness
            self.reset_display()

            new_page = self._current_page.touch_handler(e, t)
            if new_page is not None:
                if isinstance(new_page, Page):
                    self._current_page = new_page
                else:
                    LOG.error(
                        f"Unsupported touch result {type(new_page)!r}")

        # All touch types go through the `Page.touch_handler()`
        for touch in self._touchscreen.touches:
            touch.on_press = handle_touches
            touch.on_release = handle_touches
            touch.on_move = handle_touches

        try:
            while self._running:
                # Deal with screen fade and blackout
                time = pygame.time.get_ticks()
                # deltaTime in seconds.
                deltaTime = (time - self.last_event) / 1000.0
                if deltaTime > self.BLACKOUT_TIME:
                    self.set_light(False)
                    self._blacked = True
                elif deltaTime > self.FADE_TIME:
                    self.set_brightness(self.FADE_LEVEL)
                    self._faded = True

                # Render the current page
                self._current_page.render(self._touchscreen.surface)
                pygame.display.flip()

                # Check for keyboard events, to support VNC interactions
                events = pygame.event.get()
                new_page = None
                if len(events) > 0:
                    for event in events:
                        if event.type == KEYDOWN:
                            page = self._current_page
                            if event.key in [K_SPACE, K_RETURN]:
                                if page.is_focused():
                                    page.press_widget()

                            if event.key == K_UP:
                                page.shift_widget_focus(Page.GO_UP)
                            elif event.key == K_DOWN:
                                page.shift_widget_focus(Page.GO_DOWN)
                            elif event.key == K_LEFT:
                                page.shift_widget_focus(Page.GO_LEFT)
                            elif event.key == K_RIGHT:
                                page.shift_widget_focus(Page.GO_RIGHT)
                        if event.type == KEYUP:
                            page = self._current_page
                            if event.key in [K_SPACE, K_RETURN]:
                                if page.is_focused():
                                    new_page = page.release_widget()
                                    page.focus_widget(False)
                                else:
                                    page.focus_widget(True)

                            elif event.key == K_ESCAPE:
                                page.focus_widget(False)

                if pygame.key.get_pressed()[K_EQUALS]:
                    page.adjust_widget(1)
                elif pygame.key.get_pressed()[K_MINUS]:
                    page.adjust_widget(-1)

                # The keyboard interactions may have set a new page
                if new_page is not None:
                    if isinstance(new_page, Page):
                        self._current_page = new_page
                    else:
                        LOG.error(f"Unsupported page result "
                                  f"{type(new_page)!r}")

                # Lock it into the framerate
                fpsClock.tick(self.FPS)

        except Exception as e:
            backlog_error(e, f"Unmanaged exception")
        finally:
            self.reset_display()
            LOG.info("Stopping touchscreen thread...")
            self._touchscreen.stop()

            LOG.info("Exiting GUI...")
            exit()

    def stop(self):
        """ Tell the GUI to stop running.

        This method isn't really accessible, but here for completeness.
        """
        self._running = False



# Copyright (c) 2020 Edwin Wise
# MIT License
# See LICENSE for details
"""
    Manual Test - set up some pages for viewing and interaction.
"""
import logging
from pathlib import Path

import pygame

from gui import GUI
from pi_touch_gui._utilities import backlog_error
from raspberrypi_ts import Touchscreen
from test_pages.entry_page import entry_page
from test_pages.poweroff_page import poweroff_page
from test_pages.sampler_page import sampler_page

logging.basicConfig(
    level=logging.INFO,
    format="[%(asctime)s] [%(levelname)s] "
           "[%(name)s.%(lineno)d] %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S %z"
)
LOG = logging.getLogger(__name__)

if __name__ == "__main__":
    pygame.init()

    touchscreen = Touchscreen()
    interface = GUI(touchscreen)

    _entry_page = entry_page()
    _sampler_page = sampler_page()
    _poweroff_page = poweroff_page()


    # Page transition functions
    def go_to_page_1(button, pushed):
        return _entry_page


    def go_to_page_2(page, pushed):
        return _sampler_page


    def go_to_page_3(button, pushed):
        return _poweroff_page


    _entry_page.config(go_to_page_2)
    _background = str(Path(Path(__file__).parent, "assets/lcars_screen.png"))
    _sampler_page.config(_background, go_to_page_3, go_to_page_1)
    _poweroff_page.config(go_to_page_2)

    # Page Definitions

    try:
        # Add any hardware or other initialization here
        interface.run(_entry_page)
    except Exception as e:
        backlog_error(e, f"Unmanaged exception")
    finally:
        # Add any hardware or other teardown here
        pass



# Copyright (c) 2020 Edwin Wise
# MIT License
# See LICENSE for details
"""
    Testing utility functions that don't fit in any particular class or module.
"""
import math

import pygame

from colors import LCARS_PINK, LCARS_BLUE, WHITE

fake_data_1 = None
fake_data_2 = None
fake_num_samples = 1000
last_update = pygame.time.get_ticks()


def pulsing_light_func(light):
    """ Provide a dynamic light for testing.

    Specify a `rate` attribute on the Light calling me to change pulse rate.

    Parameters
    ----------
    light : Light

    Returns
    -------
    Tuple(Color, Color)
        The ring, fill colors of the light.
    """
    rate = getattr(light, 'rate')
    ring = WHITE

    time = pygame.time.get_ticks()
    red = math.sin(time / rate) + 1.0  # sets to range 0..2
    fill = (red * 127, 0, 0)

    return ring, fill


def fake_data_graph_function(graph, size):
    """ Generate two traces of randomized fake data.

    This is a Graph update function.

    Parameters
    ----------
    graph : Graph
    size : int
        The number of samples to return

    Returns
    -------
    List(Dict)
        A list of data dictionaries, where each dict takes the form:
            {"color": <Color>, "data": [<float>]}
    """
    import random
    global fake_data_1, fake_data_2
    global last_update
    global fake_num_samples

    if fake_data_1 is None:
        fake_data_1 = [5 + random.random()]
        fake_data_2 = [5 + random.random()]
    else:
        time = pygame.time.get_ticks()
        # Update twice per second, to show some life
        if (time - last_update) > 50:
            fake_data_1.append(fake_data_1[-1] + (1.0 * random.random() - 0.5))
            fake_data_2.append(fake_data_2[-1] + (1.0 * random.random() - 0.5))
            last_update = time

    # Truncate to the maximum size to prevent memory crash
    if len(fake_data_1) > fake_num_samples:
        fake_data_1 = fake_data_1[-fake_num_samples:]
        fake_data_2 = fake_data_2[-fake_num_samples:]

    # Return what was asked for
    return [{"color": LCARS_PINK, "data": fake_data_1[-size:]},
            {"color": LCARS_BLUE, "data": fake_data_2[-size:]}, ]



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="pi-touch-gui",
    version="0.1.21",
    author="Edwin Wise",
    author_email="edwin@simreal.com",
    description="A graphic UI optimized for touch screens",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/EdwinWiseOne/PiTouchGui",
    packages=setuptools.find_packages(),
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "License :: Free for non-commercial use",
        "Programming Language :: Python :: 3 :: Only",
        "Topic :: Home Automation",
        "Topic :: Software Development :: Libraries",
    ],
    python_requires='>=3.6',
    install_requires=['pygame'],
)


# Copyright (c) 2020 Edwin Wise
# MIT License
# See LICENSE for details
"""
    Test - Widget Sample, all the widgets, configured for keyboard grid.
"""

import pygame

from colors import (WHITE, MEDIUM_GRAY, BLUE, RED, BLACK, DARK_RED, GREEN, CYAN,
                    MAGENTA)
from gui import Page
from tests._utilities import fake_data_graph_function
from widget_buttons import Button, ButtonGroup
from widget_controls import Dial, Slider
from widget_displays import Indicator, Label, Graph

_boop_count = 4

_up_down_flag = 0

_max_track_samples = 1000
_last_track_update = pygame.time.get_ticks()
_track_1_data = []
_track_2_data = []
_track_3_data = []
_dial_value = 50
_slider_h_value = 40
_slider_v_value = 60


def sampler_page():
    # =========================================================
    # Hold-behavior buttons and indicator
    # =========================================================

    def _up_dn_button_func(button, pushed):
        global _up_down_flag
        direction = button.direction
        if pushed:
            _up_down_flag = direction
        else:
            _up_down_flag = 0

    def _up_dn_indicator_func(indicator):
        global _up_down_flag
        if _up_down_flag == 0:
            return False, ''
        elif _up_down_flag == 1:
            return True, "UP"
        else:
            return True, "DN"

    _up_button = Button((115, 105), (60, 50),
                        label="^",
                        behavior=Button.HOLD,
                        style=Button.OBROUND,
                        function=_up_dn_button_func)
    _up_button.direction = 1

    _up_dn_indicator = Indicator((115, 160), (60, 50),
                                 label_color2=WHITE,
                                 function=_up_dn_indicator_func)

    _dn_button = Button((115, 215), (60, 50),
                        label="v",
                        behavior=Button.HOLD,
                        style=Button.OBROUND,
                        function=_up_dn_button_func)
    _dn_button.direction = -1

    # =========================================================
    # Rounded Button styles
    # =========================================================

    _back_button = Button((110, 400), (75, 50),
                          label="Back",
                          style=Button.LEFT_ROUND)

    _do_nothing_button = Button((190, 400), (150, 50),
                                label="Do Nothing",
                                style=Button.RECTANGLE,
                                color1=MEDIUM_GRAY, color2=WHITE,
                                label_color1=BLUE, label_color2=RED)

    _next_button = Button((345, 400), (75, 50),
                          label="Next",
                          style=Button.RIGHT_ROUND)

    # ---- Boop button, for some amusement

    def _boop_button_func(button, pushed):
        global _boop_count
        _boop_count -= 1
        return None

    def _boop_indicator_func(indicator):
        """ Countdown; when we reach 0, we exit the program.
        """
        if _boop_count == 4:
            return False, ''
        if _boop_count == 0:
            exit(0)
        return True, f"{_boop_count}"

    _boop_button = Button((430, 400), (50, 50),
                          label="X",
                          style=Button.OBROUND,
                          function=_boop_button_func)

    _boop_indicator = Indicator((485, 400), (50, 50),
                                color1=BLACK, color2=RED,
                                label_color1=WHITE, label_color2=BLACK,
                                function=_boop_indicator_func)

    # =========================================================
    # Toggle-behavior Buttons with reset
    # =========================================================

    def _reset_count_buttons(button, pressed):
        _one_button.clear()
        _two_button.clear()
        _three_button.clear()

    _count_buttons_label = Label((190, 105), (100, 25),
                                 label="Pick Several")
    _one_button = Button((190, 130), (100, 50),
                         label="One",
                         behavior=Button.TOGGLE)
    _two_button = Button((190, 185), (100, 50),
                         label="Two",
                         behavior=Button.TOGGLE)
    _three_button = Button((190, 240), (100, 50),
                           label="Three",
                           behavior=Button.TOGGLE)
    _count_reset_button = Button((190, 295), (100, 50),
                                 label='RESET',
                                 label_color1=DARK_RED,
                                 function=_reset_count_buttons)

    # =========================================================
    # Button Group
    # =========================================================

    _color_buttons_label = Label((300, 105), (100, 25),
                                 label="Pick One")

    _red_button = Button((300, 130), (100, 50),
                         label="Red",
                         color2=RED,
                         behavior=Button.TOGGLE)
    _green_button = Button((300, 185), (100, 50),
                           label="Green",
                           color2=GREEN,
                           behavior=Button.TOGGLE)
    _blue_button = Button((300, 240), (100, 50), label="Blue",
                          color2=BLUE,
                          behavior=Button.TOGGLE)

    _color_group = ButtonGroup([_red_button, _green_button, _blue_button])

    # =========================================================
    # Controls and Graphs
    # =========================================================

    # ---- Graph and Control Functions

    def _dial_control_func(dial, value):
        global _dial_value
        _dial_value = value

    def _slider_h_control_func(slider, value):
        global _slider_h_value
        _slider_h_value = value

    def _slider_v_control_func(slider, value):
        global _slider_v_value
        _slider_v_value = value

    def _graph_button_update(graph, size):
        global _dial_value, _slider_h_value, _slider_v_value
        global last_update, _max_track_samples
        global _track_1_data, _track_2_data, _track_3_data
        global _last_track_update

        # Update the graph
        time = pygame.time.get_ticks()
        # Update twice per second, to show some life
        if (time - _last_track_update) > 250:
            _track_1_data.append(_dial_value)
            _track_2_data.append(_slider_h_value)
            _track_3_data.append(_slider_v_value)
            _last_track_update = time

        # Truncate to the maximum size to prevent memory crash
        if len(_track_1_data) > _max_track_samples:
            _track_1_data = _track_1_data[-_max_track_samples:]
            _track_2_data = _track_2_data[-_max_track_samples:]
            _track_3_data = _track_3_data[-_max_track_samples:]

        # Return what was asked for
        return [{"color": CYAN, "data": _track_1_data[-size:]},
                {"color": BLUE, "data": _track_2_data[-size:]},
                {"color": MAGENTA, "data": _track_3_data[-size:]}, ]

    # ---- Graph and Control Widgets

    _control_graph = Graph((410, 130), (390, 100),
                           min_max=(0, 100), function=_graph_button_update)
    _control_dial = Dial((450, 240), (150, 150),
                         min_max=(0, 100), start_value=_dial_value,
                         snap_value=1,
                         color2=CYAN,
                         dynamic=True, function=_dial_control_func)
    _h_slider = Slider((620, 400), (150, 50),
                       min_max=(25, 75), start_value=_slider_h_value,
                       snap_value=5,
                       label_color1=BLACK, color1=MEDIUM_GRAY, color2=MAGENTA,
                       dynamic=True, function=_slider_v_control_func)
    _v_slider = Slider((620, 240), (50, 150),
                       min_max=(0, 100), start_value=_slider_v_value,
                       snap_value=.1,
                       label_color1=BLACK, color1=MEDIUM_GRAY, color2=BLUE,
                       dynamic=False, function=_slider_h_control_func)
    _fake_graph = Graph((680, 240), (110, 150), min_max=(-100, 100), quiet=True,
                        function=fake_data_graph_function)

    # =========================================================
    # Define keyboard navigation grid
    # This could probably be automated someday, or bake in
    # some kind of layout manager
    # =========================================================

    # neighbors = [Up, Down, Left, Right]
    _up_button.neighbors = [Page.WALL, _dn_button, Page.WALL, _one_button]
    _dn_button.neighbors = [_up_button, _back_button, Page.WALL, _three_button]

    _back_button.neighbors = [_dn_button, Page.WALL,
                              Page.WALL, _do_nothing_button]
    _do_nothing_button.neighbors = [_count_reset_button, Page.WALL,
                                    _back_button, _next_button]
    _next_button.neighbors = [_blue_button, Page.WALL,
                              _do_nothing_button, _boop_button]

    _boop_button.neighbors = [_control_dial, Page.WALL,
                              _next_button, _h_slider]

    _one_button.neighbors = [Page.WALL, _two_button,
                             _up_button, _red_button]
    _two_button.neighbors = [_one_button, _three_button,
                             _up_button, _green_button]
    _three_button.neighbors = [_two_button, _count_reset_button,
                               _dn_button, _blue_button]
    _count_reset_button.neighbors = [_three_button, _do_nothing_button,
                                     _dn_button, _control_dial]

    _red_button.neighbors = [Page.WALL, _green_button,
                             _one_button, Page.WALL]
    _green_button.neighbors = [_red_button, _blue_button,
                               _two_button, Page.WALL]
    _blue_button.neighbors = [_green_button, _next_button,
                              _three_button, _control_dial]

    _control_dial.neighbors = [Page.WALL, _boop_button,
                               _count_reset_button, _v_slider]
    _v_slider.neighbors = [Page.WALL, _h_slider, _control_dial, Page.WALL]
    _h_slider.neighbors = [_v_slider, Page.WALL, _boop_button, Page.WALL]

    # =========================================================
    # Page configuration and definition
    # =========================================================

    def _configure_page(background, go_next, go_back):
        """ Custom configuration on the entry page - sets the page and
        button functions.

        Parameters
        ----------
        go_next : Callable[[IWidget, bool], Optional['Page']]
        go_back : Callable[[IWidget, bool], Optional['Page']]
        """
        _next_button.function = go_next
        _back_button.function = go_back
        _sampler_page.background = background

    _sampler_page = Page(
        # Tests background image, Button, RoundedButton, DynamicButton,
        # ToggleButton, ButtonGroup, Label, Indicator, ...
        [
            # Putting a non-focus element first to test that case in
            # keyboard control
            _up_dn_indicator, _up_button, _dn_button,
            _back_button, _do_nothing_button, _next_button,
            _boop_button, _boop_indicator,

            _count_buttons_label,
            _one_button, _two_button, _three_button, _count_reset_button,

            _color_buttons_label,
            _red_button, _green_button, _blue_button, _color_group,

            _control_graph, _control_dial, _v_slider, _h_slider,
            _fake_graph
        ]
    )
    # Add a custom attribute, which is monkey-patching at its worst
    _sampler_page.config = _configure_page

    return _sampler_page
